@@ -94,10 +94,10 @@ G_DEFINE_TYPE_WITH_CODE (GtkSourceCompletionWords,
 			 G_IMPLEMENT_INTERFACE (GTK_TYPE_SOURCE_COMPLETION_PROVIDER,
 				 		gtk_source_completion_words_iface_init))
 
-static gchar *
+static const gchar * 
 gtk_source_completion_words_get_name (GtkSourceCompletionProvider *self)
 {
-	return g_strdup (GTK_SOURCE_COMPLETION_WORDS (self)->priv->name);
+	return GTK_SOURCE_COMPLETION_WORDS (self)->priv->name;
 }
 
 static GdkPixbuf * 
@@ -270,9 +270,9 @@ gtk_source_completion_words_populate (GtkSourceCompletionProvider *provider,
 	}
 
 	words->priv->cancel_id = 
-		g_signal_connect_swapped (context,
-			                  "cancelled",
-			                   G_CALLBACK (population_finished),
+		g_signal_connect_swapped (context, 
+			                  "cancelled", 
+			                   G_CALLBACK (population_finished), 
 			                   provider);
 
 	words->priv->context = g_object_ref (context);
@@ -490,7 +490,7 @@ gtk_source_completion_words_class_init (GtkSourceCompletionWordsClass *klass)
 	                                                    _("Number of lines scanned in one batch"),
 	                                                    1,
 	                                                    G_MAXUINT,
-	                                                    50,
+	                                                    20,
 	                                                    G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
 	
 	g_object_class_install_property (object_class,
