@@ -1,5 +1,5 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8; coding: utf-8 -*-
- *  gtksourcebuffer.c
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8; coding: utf-8 -*- */
+/*  gtksourcebuffer.c
  *
  *  Copyright (C) 1999,2000,2001,2002 by:
  *          Mikael Hermansson <tyan@linux.se>
@@ -40,6 +40,25 @@
 #include "gtksourcestyleschememanager.h"
 #include "gtksourcestyle-private.h"
 
+/**
+ * SECTION:buffer
+ * @Short_description: Buffer object for #GtkSourceView
+ * @Title: GtkSourceBuffer
+ * @See_also: #GtkTextBuffer,#GtkSourceView
+ *
+ * The #GtkSourceBuffer object is the model for #GtkSourceView widgets.
+ * It extends the #GtkTextBuffer object by adding features useful to display
+ * and edit source code as syntax highlighting and bracket matching. It
+ * also implements support for undo/redo operations.
+ *
+ * To create a #GtkSourceBuffer use gtk_source_buffer_new() or
+ * gtk_source_buffer_new_with_language(). The second form is just a convenience
+ * function which allows you to initially set a #GtkSourceLanguage.
+ *
+ * By default highlighting is enabled, but you can disable it with
+ * gtk_source_buffer_set_highlight_syntax().
+ */
+
 /*
 #define ENABLE_DEBUG
 #define ENABLE_PROFILE
@@ -61,10 +80,14 @@
 
 #define MAX_CHARS_BEFORE_FINDING_A_MATCH    10000
 
+#define TAG_CONTEXT_CLASS_NAME "GtkSourceViewTagContextClassName"
+
 /* Signals */
 enum {
 	HIGHLIGHT_UPDATED,
 	SOURCE_MARK_UPDATED,
+	UNDO,
+	REDO,
 	LAST_SIGNAL
 };
 
@@ -128,6 +151,12 @@ static void 	 gtk_source_buffer_real_insert_text 	(GtkTextBuffer           *buff
 							 GtkTextIter             *iter,
 							 const gchar             *text,
 							 gint                     len);
+static void	 gtk_source_buffer_real_insert_pixbuf	(GtkTextBuffer           *buffer,
+							 GtkTextIter             *pos,
+							 GdkPixbuf               *pixbuf);
+static void	 gtk_source_buffer_real_insert_anchor	(GtkTextBuffer           *buffer,
+							 GtkTextIter             *pos,
+							 GtkTextChildAnchor      *anchor);
 static void 	 gtk_source_buffer_real_delete_range 	(GtkTextBuffer           *buffer,
 							 GtkTextIter             *iter,
 							 GtkTextIter             *end);
@@ -136,8 +165,12 @@ static void 	 gtk_source_buffer_real_mark_set	(GtkTextBuffer		 *buffer,
 							 GtkTextMark		 *mark);
 static void 	 gtk_source_buffer_real_mark_deleted	(GtkTextBuffer		 *buffer,
 							 GtkTextMark		 *mark);
-static gboolean	 gtk_source_buffer_find_bracket_match_with_limit (GtkTextIter    *orig,
-								  gint            max_chars);
+static gboolean	 gtk_source_buffer_find_bracket_match_with_limit (GtkSourceBuffer *buffer,
+								  GtkTextIter     *orig,
+								  gint             max_chars);
+
+static void	 gtk_source_buffer_real_undo		(GtkSourceBuffer	 *buffer);
+static void	 gtk_source_buffer_real_redo		(GtkSourceBuffer	 *buffer);
 
 static void
 gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
@@ -155,14 +188,17 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 	object_class->get_property = gtk_source_buffer_get_property;
 	object_class->set_property = gtk_source_buffer_set_property;
 
-	/* Do not set these signals handlers directly on the parent_class since
-	 * that will cause problems (a loop). */
-	tb_class->delete_range  = gtk_source_buffer_real_delete_range;
-	tb_class->insert_text 	= gtk_source_buffer_real_insert_text;
+	tb_class->delete_range        = gtk_source_buffer_real_delete_range;
+	tb_class->insert_text 	      = gtk_source_buffer_real_insert_text;
+	tb_class->insert_pixbuf       = gtk_source_buffer_real_insert_pixbuf;
+	tb_class->insert_child_anchor = gtk_source_buffer_real_insert_anchor;
 
 	tb_class->mark_set	= gtk_source_buffer_real_mark_set;
 	tb_class->mark_deleted	= gtk_source_buffer_real_mark_deleted;
 
+	klass->undo = gtk_source_buffer_real_undo;
+	klass->redo = gtk_source_buffer_real_redo;
+
 	/**
 	 * GtkSourceBuffer:highlight-syntax:
 	 *
@@ -277,6 +313,26 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 			   G_TYPE_NONE,
 			   1, GTK_TYPE_TEXT_MARK);
 
+	buffer_signals[UNDO] =
+	    g_signal_new ("undo",
+			  G_OBJECT_CLASS_TYPE (object_class),
+			  G_SIGNAL_RUN_LAST,
+			  G_STRUCT_OFFSET (GtkSourceBufferClass, undo),
+			  NULL, NULL,
+			  g_cclosure_marshal_VOID__VOID,
+			  G_TYPE_NONE,
+			  0);
+
+	buffer_signals[REDO] =
+	    g_signal_new ("redo",
+			  G_OBJECT_CLASS_TYPE (object_class),
+			  G_SIGNAL_RUN_LAST,
+			  G_STRUCT_OFFSET (GtkSourceBufferClass, redo),
+			  NULL, NULL,
+			  g_cclosure_marshal_VOID__VOID,
+			  G_TYPE_NONE,
+			  0);
+
 	g_type_class_add_private (object_class, sizeof(GtkSourceBufferPrivate));
 }
 
@@ -572,6 +628,17 @@ get_bracket_match_tag (GtkSourceBuffer *buffer)
 	return buffer->priv->bracket_match_tag;
 }
 
+/*
+ * This is private, just used by the compositor to not print bracket
+ * matches. Note that unlike get_bracket_match_tag() it returns NULL
+ * if the tag is not set.
+ */
+GtkTextTag *
+_gtk_source_buffer_get_bracket_match_tag (GtkSourceBuffer *buffer)
+{
+	return buffer->priv->bracket_match_tag;
+}
+
 static void
 gtk_source_buffer_move_cursor (GtkTextBuffer     *buffer,
 			       const GtkTextIter *iter,
@@ -604,7 +671,9 @@ gtk_source_buffer_move_cursor (GtkTextBuffer     *buffer,
 		return;
 
 	iter1 = *iter;
-	if (gtk_source_buffer_find_bracket_match_with_limit (&iter1, MAX_CHARS_BEFORE_FINDING_A_MATCH))
+	if (gtk_source_buffer_find_bracket_match_with_limit (GTK_SOURCE_BUFFER (buffer),
+	                                                     &iter1,
+	                                                     MAX_CHARS_BEFORE_FINDING_A_MATCH))
 	{
 		if (!GTK_SOURCE_BUFFER (buffer)->priv->bracket_mark)
 			GTK_SOURCE_BUFFER (buffer)->priv->bracket_mark =
@@ -631,16 +700,32 @@ gtk_source_buffer_move_cursor (GtkTextBuffer     *buffer,
 	}
 }
 
+static void
+gtk_source_buffer_content_inserted (GtkTextBuffer *buffer,
+				    gint           start_offset,
+				    gint           end_offset)
+{
+	GtkTextMark *mark;
+	GtkTextIter insert_iter;
+	GtkSourceBuffer *source_buffer = GTK_SOURCE_BUFFER (buffer);
+
+	mark = gtk_text_buffer_get_insert (buffer);
+	gtk_text_buffer_get_iter_at_mark (buffer, &insert_iter, mark);
+	gtk_source_buffer_move_cursor (buffer, &insert_iter, mark);
+
+	if (source_buffer->priv->highlight_engine != NULL)
+		_gtk_source_engine_text_inserted (source_buffer->priv->highlight_engine,
+						  start_offset,
+						  end_offset);
+}
+
 static void
 gtk_source_buffer_real_insert_text (GtkTextBuffer *buffer,
 				    GtkTextIter   *iter,
 				    const gchar   *text,
 				    gint           len)
 {
-	GtkTextMark *mark;
-	GtkTextIter insert_iter;
-	gint start_offset, end_offset;
-	GtkSourceBuffer *source_buffer = GTK_SOURCE_BUFFER (buffer);
+	gint start_offset;
 
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
 	g_return_if_fail (iter != NULL);
@@ -657,16 +742,64 @@ gtk_source_buffer_real_insert_text (GtkTextBuffer *buffer,
 	 */
 	GTK_TEXT_BUFFER_CLASS (gtk_source_buffer_parent_class)->insert_text (buffer, iter, text, len);
 
-	mark = gtk_text_buffer_get_insert (buffer);
-	gtk_text_buffer_get_iter_at_mark (buffer, &insert_iter, mark);
-	gtk_source_buffer_move_cursor (buffer, &insert_iter, mark);
+	gtk_source_buffer_content_inserted (buffer,
+					    start_offset,
+					    gtk_text_iter_get_offset (iter));
+}
 
-	end_offset = gtk_text_iter_get_offset (iter);
+/* insert_pixbuf and insert_child_anchor do nothing except notifying
+ * the highlighting engine about the change, because engine's idea
+ * of buffer char count must be correct at all times */
+static void
+gtk_source_buffer_real_insert_pixbuf (GtkTextBuffer *buffer,
+				      GtkTextIter   *iter,
+				      GdkPixbuf     *pixbuf)
+{
+	gint start_offset;
 
-	if (source_buffer->priv->highlight_engine != NULL)
-		_gtk_source_engine_text_inserted (source_buffer->priv->highlight_engine,
-						  start_offset,
-						  end_offset);
+	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
+	g_return_if_fail (iter != NULL);
+	g_return_if_fail (gtk_text_iter_get_buffer (iter) == buffer);
+
+	start_offset = gtk_text_iter_get_offset (iter);
+
+	/*
+	 * iter is invalidated when
+	 * insertion occurs (because the buffer contents change), but the
+	 * default signal handler revalidates it to point to the end of the
+	 * inserted text
+	 */
+	GTK_TEXT_BUFFER_CLASS (gtk_source_buffer_parent_class)->insert_pixbuf (buffer, iter, pixbuf);
+
+	gtk_source_buffer_content_inserted (buffer,
+					    start_offset,
+					    gtk_text_iter_get_offset (iter));
+}
+
+static void
+gtk_source_buffer_real_insert_anchor (GtkTextBuffer      *buffer,
+				      GtkTextIter        *iter,
+				      GtkTextChildAnchor *anchor)
+{
+	gint start_offset;
+
+	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
+	g_return_if_fail (iter != NULL);
+	g_return_if_fail (gtk_text_iter_get_buffer (iter) == buffer);
+
+	start_offset = gtk_text_iter_get_offset (iter);
+
+	/*
+	 * iter is invalidated when
+	 * insertion occurs (because the buffer contents change), but the
+	 * default signal handler revalidates it to point to the end of the
+	 * inserted text
+	 */
+	GTK_TEXT_BUFFER_CLASS (gtk_source_buffer_parent_class)->insert_child_anchor (buffer, iter, anchor);
+
+	gtk_source_buffer_content_inserted (buffer,
+					    start_offset,
+					    gtk_text_iter_get_offset (iter));
 }
 
 static void
@@ -701,35 +834,36 @@ gtk_source_buffer_real_delete_range (GtkTextBuffer *buffer,
 						 offset, length);
 }
 
-/* FIXME: this can't be here, but it's now needed for bracket matching */
-static gpointer
-iter_has_syntax_tag (const GtkTextIter *iter)
+/* This describes a mask of relevant context classes for highlighting matching
+   brackets. Additional classes can be added below */
+static const gchar *cclass_mask_definitions[] = {
+	"comment",
+	"string",
+};
+
+static gint
+get_context_class_mask (GtkSourceBuffer *buffer,
+                        GtkTextIter     *iter)
 {
-	return NULL;
-// 	const GtkSourceTag *tag;
-// 	GSList *list;
-// 	GSList *l;
-//
-// 	g_return_val_if_fail (iter != NULL, NULL);
-//
-// 	list = gtk_text_iter_get_tags (iter);
-// 	tag = NULL;
-//
-// 	l = list;
-//
-// 	while ((list != NULL) && (tag == NULL)) {
-// 		if (GTK_IS_SOURCE_TAG (list->data))
-// 			tag = GTK_SOURCE_TAG (list->data);
-// 		list = g_slist_next (list);
-// 	}
-//
-// 	g_slist_free (l);
-//
-// 	return tag;
+	gint i;
+	gint ret = 0;
+
+	for (i = 0; i < sizeof (cclass_mask_definitions) / sizeof (gchar *); ++i)
+	{
+		gboolean hasclass = gtk_source_buffer_iter_has_context_class (buffer,
+		                                                              iter,
+		                                                              cclass_mask_definitions[i]);
+
+		ret |= hasclass << i;
+	}
+
+	return ret;
 }
 
 static gboolean
-gtk_source_buffer_find_bracket_match_real (GtkTextIter *orig, gint max_chars)
+gtk_source_buffer_find_bracket_match_real (GtkSourceBuffer *buffer,
+                                           GtkTextIter     *orig,
+                                           gint             max_chars)
 {
 	GtkTextIter iter;
 
@@ -742,14 +876,14 @@ gtk_source_buffer_find_bracket_match_real (GtkTextIter *orig, gint max_chars)
 
 	gboolean found;
 
-	gpointer base_tag;
+	gint cclass_mask;
 
 	iter = *orig;
 
 	cur_char = gtk_text_iter_get_char (&iter);
 
 	base_char = search_char = cur_char;
-	base_tag = iter_has_syntax_tag (&iter);
+	cclass_mask = get_context_class_mask (buffer, &iter);
 
 	switch ((int) base_char) {
 		case '{':
@@ -797,12 +931,24 @@ gtk_source_buffer_find_bracket_match_real (GtkTextIter *orig, gint max_chars)
 	char_cont = 0;
 
 	do {
+		gint current_mask;
+
 		gtk_text_iter_forward_chars (&iter, addition);
 		cur_char = gtk_text_iter_get_char (&iter);
 		++char_cont;
 
+		current_mask = get_context_class_mask (buffer, &iter);
+
+		/* Check if we lost a class, which means we don't look any
+		   further */
+		if (current_mask < cclass_mask)
+		{
+			found = FALSE;
+			break;
+		}
+
 		if ((cur_char == search_char || cur_char == base_char) &&
-		    base_tag == iter_has_syntax_tag (&iter))
+		    cclass_mask == current_mask)
 		{
 			if ((cur_char == search_char) && counter == 0) {
 				found = TRUE;
@@ -828,11 +974,13 @@ gtk_source_buffer_find_bracket_match_real (GtkTextIter *orig, gint max_chars)
  * cursor takes precedence.
  */
 static gboolean
-gtk_source_buffer_find_bracket_match_with_limit (GtkTextIter *orig, gint max_chars)
+gtk_source_buffer_find_bracket_match_with_limit (GtkSourceBuffer *buffer,
+                                                 GtkTextIter     *orig,
+                                                 gint             max_chars)
 {
 	GtkTextIter iter;
 
-	if (gtk_source_buffer_find_bracket_match_real (orig, max_chars))
+	if (gtk_source_buffer_find_bracket_match_real (buffer, orig, max_chars))
 	{
 		return TRUE;
 	}
@@ -841,7 +989,7 @@ gtk_source_buffer_find_bracket_match_with_limit (GtkTextIter *orig, gint max_cha
 	if (!gtk_text_iter_starts_line (&iter) &&
 	    gtk_text_iter_backward_char (&iter))
 	{
-		if (gtk_source_buffer_find_bracket_match_real (&iter, max_chars))
+		if (gtk_source_buffer_find_bracket_match_real (buffer, &iter, max_chars))
 		{
 			*orig = iter;
 			return TRUE;
@@ -924,9 +1072,8 @@ void
 gtk_source_buffer_undo (GtkSourceBuffer *buffer)
 {
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
-	g_return_if_fail (gtk_source_undo_manager_can_undo (buffer->priv->undo_manager));
 
-	gtk_source_undo_manager_undo (buffer->priv->undo_manager);
+	g_signal_emit (buffer, buffer_signals[UNDO], 0);
 }
 
 /**
@@ -940,9 +1087,8 @@ void
 gtk_source_buffer_redo (GtkSourceBuffer *buffer)
 {
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
-	g_return_if_fail (gtk_source_undo_manager_can_redo (buffer->priv->undo_manager));
 
-	gtk_source_undo_manager_redo (buffer->priv->undo_manager);
+	g_signal_emit (buffer, buffer_signals[REDO], 0);
 }
 
 /**
@@ -1155,6 +1301,7 @@ gtk_source_buffer_set_language (GtkSourceBuffer   *buffer,
 				GtkSourceLanguage *language)
 {
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
+	g_return_if_fail (GTK_IS_SOURCE_LANGUAGE (language) || language == NULL);
 
 	if (buffer->priv->language == language)
 		return;
@@ -1267,14 +1414,15 @@ gtk_source_buffer_ensure_highlight (GtkSourceBuffer   *buffer,
  * @buffer: a #GtkSourceBuffer.
  * @scheme: style scheme.
  *
- * Sets style scheme used by the buffer.
+ * Sets style scheme used by the buffer. If @scheme is %NULL no
+ * style scheme is used.
  **/
 void
 gtk_source_buffer_set_style_scheme (GtkSourceBuffer      *buffer,
 				    GtkSourceStyleScheme *scheme)
 {
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
-	g_return_if_fail (GTK_IS_SOURCE_STYLE_SCHEME (scheme));
+	g_return_if_fail (GTK_IS_SOURCE_STYLE_SCHEME (scheme) || scheme == NULL);
 
 	if (buffer->priv->style_scheme == scheme)
 		return;
@@ -1282,7 +1430,7 @@ gtk_source_buffer_set_style_scheme (GtkSourceBuffer      *buffer,
 	if (buffer->priv->style_scheme)
 		g_object_unref (buffer->priv->style_scheme);
 
-	buffer->priv->style_scheme = g_object_ref (scheme);
+	buffer->priv->style_scheme = scheme ? g_object_ref (scheme) : NULL;
 	update_bracket_match_style (buffer);
 
 	if (buffer->priv->highlight_engine != NULL)
@@ -1317,7 +1465,7 @@ gtk_source_buffer_get_style_scheme (GtkSourceBuffer *buffer)
 static gboolean
 source_mark_remove (GtkSourceBuffer *buffer, GtkSourceMark *mark)
 {
-	gint i;
+	guint i;
 
 	for (i = 0; i < buffer->priv->source_marks->len; ++i)
 	{
@@ -1389,27 +1537,27 @@ static void
 source_mark_insert (GtkSourceBuffer *buffer, GtkSourceMark *mark)
 {
 	GtkTextIter iter;
-	gint index, cmp;
+	gint idx, cmp;
 
 	gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (buffer),
 					  &iter,
 					  GTK_TEXT_MARK (mark));
 
-	index = source_mark_bsearch (buffer, &iter, &cmp);
-	if (index >= 0)
+	idx = source_mark_bsearch (buffer, &iter, &cmp);
+	if (idx >= 0)
 	{
 		/* if the mark we found is at same iter or before
 		 * put our mark after that */
 		if (cmp >= 0)
-			index++;
+			idx++;
 	}
 	else
 	{
-		index = 0;
+		idx = 0;
 	}
 
 	g_object_ref (mark);
-	g_array_insert_val (buffer->priv->source_marks, index, mark);
+	g_array_insert_val (buffer->priv->source_marks, idx, mark);
 }
 
 static void
@@ -1454,6 +1602,22 @@ gtk_source_buffer_real_mark_deleted (GtkTextBuffer *buffer,
 		GTK_TEXT_BUFFER_CLASS (gtk_source_buffer_parent_class)->mark_deleted (buffer, mark);
 }
 
+static void
+gtk_source_buffer_real_undo (GtkSourceBuffer *buffer)
+{
+	g_return_if_fail (gtk_source_undo_manager_can_undo (buffer->priv->undo_manager));
+
+	gtk_source_undo_manager_undo (buffer->priv->undo_manager);
+}
+
+static void
+gtk_source_buffer_real_redo (GtkSourceBuffer *buffer)
+{
+	g_return_if_fail (gtk_source_undo_manager_can_redo (buffer->priv->undo_manager));
+
+	gtk_source_undo_manager_redo (buffer->priv->undo_manager);
+}
+
 /**
  * gtk_source_buffer_create_source_mark:
  * @buffer: a #GtkSourceBuffer.
@@ -1505,7 +1669,7 @@ _gtk_source_buffer_source_mark_next (GtkSourceBuffer *buffer,
 				     const gchar     *category)
 {
 	GtkTextIter iter;
-	gint index, cmp;
+	gint idx, cmp;
 
 	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), NULL);
 
@@ -1516,23 +1680,23 @@ _gtk_source_buffer_source_mark_next (GtkSourceBuffer *buffer,
 					  &iter,
 					  GTK_TEXT_MARK (mark));
 
-	index = source_mark_bsearch (buffer, &iter, &cmp);
+	idx = source_mark_bsearch (buffer, &iter, &cmp);
 
 	/* the array should already contain @mark */
-	g_return_val_if_fail (index >= 0, NULL);
+	g_return_val_if_fail (idx >= 0, NULL);
 	g_return_val_if_fail (cmp == 0, NULL);
 
 	/* move up to our mark among the ones at this position */
-	while (mark != g_array_index (buffer->priv->source_marks, GtkSourceMark *, index))
+	while (mark != g_array_index (buffer->priv->source_marks, GtkSourceMark *, idx))
 	{
-		++index;
+		++idx;
 	}
 
-	while (++index < buffer->priv->source_marks->len)
+	while ((guint) ++idx < buffer->priv->source_marks->len)
 	{
 		GtkSourceMark *ret;
 
-		ret = g_array_index (buffer->priv->source_marks, GtkSourceMark *, index);
+		ret = g_array_index (buffer->priv->source_marks, GtkSourceMark *, idx);
 		if (category == NULL ||
 		    0 == strcmp (category, gtk_source_mark_get_category (ret)))
 		{
@@ -1549,7 +1713,7 @@ _gtk_source_buffer_source_mark_prev (GtkSourceBuffer *buffer,
 				     const gchar     *category)
 {
 	GtkTextIter iter;
-	gint index, cmp;
+	gint idx, cmp;
 
 	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), NULL);
 
@@ -1560,23 +1724,23 @@ _gtk_source_buffer_source_mark_prev (GtkSourceBuffer *buffer,
 					  &iter,
 					  GTK_TEXT_MARK (mark));
 
-	index = source_mark_bsearch (buffer, &iter, &cmp);
+	idx = source_mark_bsearch (buffer, &iter, &cmp);
 
 	/* the array should already contain @mark */
-	g_return_val_if_fail (index >= 0, NULL);
+	g_return_val_if_fail (idx >= 0, NULL);
 	g_return_val_if_fail (cmp == 0, NULL);
 
 	/* move up to our mark among the ones at this position */
-	while (mark != g_array_index (buffer->priv->source_marks, GtkSourceMark *, index))
+	while (mark != g_array_index (buffer->priv->source_marks, GtkSourceMark *, idx))
 	{
-		++index;
+		++idx;
 	}
 
-	while (--index >= 0)
+	while (--idx >= 0)
 	{
 		GtkSourceMark *ret;
 
-		ret = g_array_index (buffer->priv->source_marks, GtkSourceMark *, index);
+		ret = g_array_index (buffer->priv->source_marks, GtkSourceMark *, idx);
 		if (category == NULL ||
 		    0 == strcmp (category, gtk_source_mark_get_category (ret)))
 		{
@@ -1607,25 +1771,25 @@ gtk_source_buffer_forward_iter_to_source_mark (GtkSourceBuffer *buffer,
 					       const gchar     *category)
 {
 	GtkTextIter i;
-	gint index, cmp;
+	gint idx, cmp;
 
 	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), FALSE);
 	g_return_val_if_fail (iter != NULL, FALSE);
 
 	i = *iter;
 
-	index = source_mark_bsearch (buffer, &i, &cmp);
-	if (index < 0)
+	idx = source_mark_bsearch (buffer, &i, &cmp);
+	if (idx < 0)
 		return FALSE;
 
 	if (cmp >= 0)
-		++index;
+		++idx;
 
-	while (index < buffer->priv->source_marks->len)
+	while ((guint) idx < buffer->priv->source_marks->len)
 	{
 		GtkSourceMark *mark;
 
-		mark = g_array_index (buffer->priv->source_marks, GtkSourceMark *, index);
+		mark = g_array_index (buffer->priv->source_marks, GtkSourceMark *, idx);
 		if (category == NULL ||
 		    0 == strcmp (category, gtk_source_mark_get_category (mark)))
 		{
@@ -1640,7 +1804,7 @@ gtk_source_buffer_forward_iter_to_source_mark (GtkSourceBuffer *buffer,
 			}
 		}
 
-		++index;
+		++idx;
 	}
 
 	return FALSE;
@@ -1666,25 +1830,25 @@ gtk_source_buffer_backward_iter_to_source_mark (GtkSourceBuffer *buffer,
 						const gchar     *category)
 {
 	GtkTextIter i;
-	gint index, cmp;
+	gint idx, cmp;
 
 	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), FALSE);
 	g_return_val_if_fail (iter != NULL, FALSE);
 
 	i = *iter;
 
-	index = source_mark_bsearch (buffer, &i, &cmp);
-	if (index < 0)
+	idx = source_mark_bsearch (buffer, &i, &cmp);
+	if (idx < 0)
 		return FALSE;
 
 	if (cmp <= 0)
-		--index;
+		--idx;
 
-	while (index >= 0)
+	while (idx >= 0)
 	{
 		GtkSourceMark *mark;
 
-		mark = g_array_index (buffer->priv->source_marks, GtkSourceMark *, index);
+		mark = g_array_index (buffer->priv->source_marks, GtkSourceMark *, idx);
 		if (category == NULL ||
 		    0 == strcmp (category, gtk_source_mark_get_category (mark)))
 		{
@@ -1698,7 +1862,7 @@ gtk_source_buffer_backward_iter_to_source_mark (GtkSourceBuffer *buffer,
 			}
 		}
 
-		--index;
+		--idx;
 	}
 
 	return FALSE;
@@ -1859,3 +2023,174 @@ gtk_source_buffer_remove_source_marks (GtkSourceBuffer   *buffer,
 	g_slist_free (list);
 }
 
+
+/**
+ * gtk_source_buffer_iter_has_context_class:
+ * @buffer: a #GtkSourceBuffer.
+ * @iter: a #GtkTextIter
+ * @context_class: class to search for
+ *
+ * Check if the class @context_klass is set on @iter.
+ *
+ * Since: 2.10
+ **/
+gboolean
+gtk_source_buffer_iter_has_context_class (GtkSourceBuffer   *buffer,
+                                          const GtkTextIter *iter,
+                                          const gchar       *context_class)
+{
+	GtkTextTag *tag;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), FALSE);
+	g_return_val_if_fail (iter != NULL, FALSE);
+	g_return_val_if_fail (context_class != NULL, FALSE);
+
+	if (buffer->priv->highlight_engine == NULL)
+	{
+		return FALSE;
+	}
+
+	tag = _gtk_source_engine_get_context_class_tag (buffer->priv->highlight_engine,
+							context_class);
+
+	if (tag != NULL)
+	{
+		return gtk_text_iter_has_tag (iter, tag);
+	}
+	else
+	{
+		return FALSE;
+	}
+}
+
+/**
+ * gtk_source_buffer_get_context_classes_at_iter:
+ * @buffer: a #GtkSourceBuffer.
+ * @iter: a #GtkTextIter
+ *
+ * Get all defined context classes at @iter.
+ *
+ * Returns: a new %NULL terminated array of context class names. Use
+ *          #g_strfreev to free the array if it is no longer needed.
+ *
+ * Since: 2.10
+ **/
+gchar **
+gtk_source_buffer_get_context_classes_at_iter (GtkSourceBuffer   *buffer,
+                                               const GtkTextIter *iter)
+{
+	GSList *tags;
+	GSList *item;
+	GPtrArray *ret;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), NULL);
+	g_return_val_if_fail (iter != NULL, NULL);
+
+	tags = gtk_text_iter_get_tags (iter);
+	ret = g_ptr_array_new ();
+
+	for (item = tags; item; item = g_slist_next (item))
+	{
+		gchar const *name = g_object_get_data (G_OBJECT (item->data),
+		                                       TAG_CONTEXT_CLASS_NAME);
+
+		if (name != NULL)
+		{
+			g_ptr_array_add (ret, g_strdup (name));
+		}
+	}
+
+	g_ptr_array_add (ret, NULL);
+	return (gchar **) g_ptr_array_free (ret, FALSE);
+}
+
+/**
+ * gtk_source_buffer_iter_forward_to_context_class_toggle:
+ * @buffer: a #GtkSourceBuffer.
+ * @iter: a #GtkTextIter
+ * @context_class: the context class
+ *
+ * Moves forward to the next toggle (on or off) of the context class. If no
+ * matching context class toggles are found, returns %FALSE, otherwise %TRUE.
+ * Does not return toggles located at @iter, only toggles after @iter. Sets
+ * @iter to the location of the toggle, or to the end of the buffer if no
+ * toggle is found.
+ *
+ * Returns: whether we found a context class toggle after @iter
+ *
+ * Since: 2.10
+ **/
+gboolean
+gtk_source_buffer_iter_forward_to_context_class_toggle (GtkSourceBuffer *buffer,
+                                                        GtkTextIter     *iter,
+                                                        const gchar     *context_class)
+{
+	GtkTextTag *tag;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), FALSE);
+	g_return_val_if_fail (iter != NULL, FALSE);
+	g_return_val_if_fail (context_class != NULL, FALSE);
+
+	if (buffer->priv->highlight_engine == NULL)
+	{
+		return FALSE;
+	}
+
+	tag = _gtk_source_engine_get_context_class_tag (buffer->priv->highlight_engine,
+							context_class);
+
+	if (tag == NULL)
+	{
+		return FALSE;
+	}
+	else
+	{
+		return gtk_text_iter_forward_to_tag_toggle (iter, tag);
+	}
+}
+
+/**
+ * gtk_source_buffer_iter_backward_to_context_class_toggle:
+ * @buffer: a #GtkSourceBuffer.
+ * @iter: a #GtkTextIter
+ * @context_class: the context class
+ *
+ * Moves backward to the next toggle (on or off) of the context class. If no
+ * matching context class toggles are found, returns %FALSE, otherwise %TRUE.
+ * Does not return toggles located at @iter, only toggles after @iter. Sets
+ * @iter to the location of the toggle, or to the end of the buffer if no
+ * toggle is found.
+ *
+ * Returns: whether we found a context class toggle before @iter
+ *
+ * Since: 2.10
+ **/
+gboolean
+gtk_source_buffer_iter_backward_to_context_class_toggle (GtkSourceBuffer *buffer,
+                                                         GtkTextIter     *iter,
+                                                         const gchar     *context_class)
+{
+	GtkTextTag *tag;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), FALSE);
+	g_return_val_if_fail (iter != NULL, FALSE);
+	g_return_val_if_fail (context_class != NULL, FALSE);
+
+	if (buffer->priv->highlight_engine == NULL)
+	{
+		return FALSE;
+	}
+
+	tag = _gtk_source_engine_get_context_class_tag (buffer->priv->highlight_engine,
+							context_class);
+
+	if (tag == NULL)
+	{
+		return FALSE;
+	}
+	else
+	{
+		return gtk_text_iter_backward_to_tag_toggle (iter, tag);
+	}
+}
+
