@@ -55,13 +55,15 @@ struct _GtkSourceBufferClass
 {
 	GtkTextBufferClass parent_class;
 
+	/* Signals */
+	void (*undo) (GtkSourceBuffer *buffer);
+	void (*redo) (GtkSourceBuffer *buffer);
+
 	/* Padding for future expansion */
 	void (*_gtk_source_reserved1) (void);
 	void (*_gtk_source_reserved2) (void);
 	void (*_gtk_source_reserved3) (void);
 	void (*_gtk_source_reserved4) (void);
-	void (*_gtk_source_reserved5) (void);
-	void (*_gtk_source_reserved6) (void);
 };
 
 GType           	 gtk_source_buffer_get_type 		(void) G_GNUC_CONST;
@@ -134,6 +136,26 @@ void			 gtk_source_buffer_remove_source_marks	(GtkSourceBuffer        *buffer,
 								 const GtkTextIter      *start,
 								 const GtkTextIter      *end,
 								 const gchar            *category);
+
+gboolean		 gtk_source_buffer_iter_has_context_class
+								(GtkSourceBuffer	*buffer,
+								 const GtkTextIter	*iter,
+								 const gchar            *context_class);
+
+gchar			**gtk_source_buffer_get_context_classes_at_iter
+								(GtkSourceBuffer	*buffer,
+								 const GtkTextIter	*iter);
+
+gboolean		 gtk_source_buffer_iter_forward_to_context_class_toggle
+								(GtkSourceBuffer	*buffer,
+								 GtkTextIter		*iter,
+								 const gchar		*context_class);
+
+gboolean		 gtk_source_buffer_iter_backward_to_context_class_toggle
+								(GtkSourceBuffer	*buffer,
+								 GtkTextIter		*iter,
+								 const gchar		*context_class);
+
 /* private */
 void			 _gtk_source_buffer_update_highlight	(GtkSourceBuffer        *buffer,
 								 const GtkTextIter      *start,
@@ -146,6 +168,9 @@ GtkSourceMark		*_gtk_source_buffer_source_mark_next	(GtkSourceBuffer        *buf
 GtkSourceMark		*_gtk_source_buffer_source_mark_prev	(GtkSourceBuffer        *buffer,
 								 GtkSourceMark          *mark, 
 								 const gchar            *category);
+
+GtkTextTag		*_gtk_source_buffer_get_bracket_match_tag (GtkSourceBuffer        *buffer);
+
 G_END_DECLS
 
 #endif /* __GTK_SOURCE_BUFFER_H__ */
