@@ -2,7 +2,7 @@
  * gtksourcecompletioncontext.c
  * This file is part of gtksourceview
  *
- * Copyright (C) 2009 - Jesse van den Kieboom
+ * Copyright (C) 2009 - Jesse van den Kieboom <jessevdk@gnome.org>
  *
  * gtksourceview is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -31,17 +31,18 @@ struct _GtkSourceCompletionContextPrivate
 {
 	GtkSourceCompletion *completion;
 
-	GtkTextIter iter;
+	GtkTextMark *mark;
 	GtkSourceCompletionActivation activation;
+
+	gulong mark_set_id;
 };
 
 /* Properties */
 enum
 {
 	PROP_0,
-	
+
 	PROP_COMPLETION,
-	PROP_VIEW,
 	PROP_ITER,
 	PROP_ACTIVATION
 };
@@ -57,38 +58,83 @@ guint context_signals[NUM_SIGNALS] = {0,};
 
 G_DEFINE_TYPE (GtkSourceCompletionContext, gtk_source_completion_context, G_TYPE_INITIALLY_UNOWNED)
 
+/* FIXME: we use this util to get the buffer from the completion
+   but this object is not robust to a change of the buffer associated
+   to the view. Context lifetime should be short enough to not really
+   matter.
+*/
+
+static GtkTextBuffer *
+get_buffer (GtkSourceCompletionContext *context)
+{
+	GtkSourceView *view = gtk_source_completion_get_view (context->priv->completion);
+	return gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
+}
+
 static void
 gtk_source_completion_context_dispose (GObject *object)
 {
-	GtkSourceCompletionContext *context = GTK_SOURCE_COMPLETION_CONTEXT (object);
-	
+	GtkSourceCompletionContext *context;
+	GtkTextBuffer *buffer;
+
+	context = GTK_SOURCE_COMPLETION_CONTEXT (object);
+	buffer = get_buffer (context);
+
+	if (context->priv->mark_set_id)
+	{
+		g_signal_handler_disconnect (buffer, context->priv->mark_set_id);
+		context->priv->mark_set_id = 0;
+	}
+
 	if (context->priv->completion)
 	{
 		g_object_unref (context->priv->completion);
 		context->priv->completion = NULL;
 	}
 
+	if (context->priv->mark)
+	{
+		gtk_text_buffer_delete_mark (buffer, context->priv->mark);
+		context->priv->mark = NULL;
+	}
+
 	G_OBJECT_CLASS (gtk_source_completion_context_parent_class)->dispose (object);
 }
 
+static void
+gtk_source_completion_context_set_iter (GtkSourceCompletionContext *context,
+                                        GtkTextIter                *iter)
+{
+	GtkTextBuffer *buffer = get_buffer (context);
+
+	if (context->priv->mark == NULL)
+	{
+		context->priv->mark = gtk_text_buffer_create_mark (buffer, NULL, iter, FALSE);
+	}
+	else
+	{
+		gtk_text_buffer_move_mark (buffer, context->priv->mark, iter);
+	}
+}
+
 static void
 gtk_source_completion_context_set_property (GObject      *object,
                                             guint         prop_id,
                                             const GValue *value,
                                             GParamSpec   *pspec)
 {
-	GtkSourceCompletionContext *self = GTK_SOURCE_COMPLETION_CONTEXT (object);
-	
+	GtkSourceCompletionContext *context = GTK_SOURCE_COMPLETION_CONTEXT (object);
+
 	switch (prop_id)
 	{
 		case PROP_COMPLETION:
-			self->priv->completion = g_value_dup_object (value);
+			context->priv->completion = g_value_dup_object (value);
 			break;
 		case PROP_ITER:
-			self->priv->iter = *((GtkTextIter *)g_value_get_pointer (value));
+			gtk_source_completion_context_set_iter (context, (GtkTextIter *) g_value_get_boxed (value));
 			break;
 		case PROP_ACTIVATION:
-			self->priv->activation = g_value_get_flags (value);
+			context->priv->activation = g_value_get_flags (value);
 			break;
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -101,27 +147,64 @@ gtk_source_completion_context_get_property (GObject    *object,
                                             GValue     *value,
                                             GParamSpec *pspec)
 {
-	GtkSourceCompletionContext *self = GTK_SOURCE_COMPLETION_CONTEXT (object);
-	
+	GtkSourceCompletionContext *context = GTK_SOURCE_COMPLETION_CONTEXT (object);
+
 	switch (prop_id)
 	{
 		case PROP_COMPLETION:
-			g_value_set_object (value, self->priv->completion);
-			break;
-		case PROP_VIEW:
-			g_value_set_object (value, gtk_source_completion_get_view (self->priv->completion));
+			g_value_set_object (value, context->priv->completion);
 			break;
 		case PROP_ITER:
-			g_value_set_pointer (value, &(self->priv->iter));
+			{
+				GtkTextIter iter;
+				gtk_source_completion_context_get_iter (context, &iter);
+				g_value_set_boxed (value, &iter);
+			}
 			break;
 		case PROP_ACTIVATION:
-			g_value_set_flags (value, self->priv->activation);
+			g_value_set_flags (value, context->priv->activation);
 			break;
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 	}
 }
 
+static void
+buffer_mark_set_cb (GtkTextBuffer              *buffer,
+                    GtkTextIter                *iter,
+                    GtkTextMark                *mark,
+                    GtkSourceCompletionContext *context)
+{
+	if (mark == context->priv->mark)
+	{
+		g_object_notify (G_OBJECT (buffer), "iter");
+	}
+}
+
+static GObject *
+gtk_source_completion_context_constructor (GType                  type,
+                                           guint                  n_construct_properties,
+                                           GObjectConstructParam *construct_param)
+{
+	GObject *object;
+	GtkTextBuffer *buffer;
+	GtkSourceCompletionContext *context;
+
+	object = G_OBJECT_CLASS (gtk_source_completion_context_parent_class)->constructor (type,
+											   n_construct_properties,
+									 		   construct_param);
+
+	/* we need to connect after the completion property is set */
+	context = GTK_SOURCE_COMPLETION_CONTEXT (object);
+	buffer = get_buffer (context);
+	context->priv->mark_set_id = g_signal_connect (buffer,
+						       "mark-set",
+						       G_CALLBACK (buffer_mark_set_cb),
+						       context);
+
+	return object;
+}
+
 static void
 gtk_source_completion_context_class_init (GtkSourceCompletionContextClass *klass)
 {
@@ -129,6 +212,7 @@ gtk_source_completion_context_class_init (GtkSourceCompletionContextClass *klass
 	
 	object_class->set_property = gtk_source_completion_context_set_property;
 	object_class->get_property = gtk_source_completion_context_get_property;
+	object_class->constructor = gtk_source_completion_context_constructor;
 	object_class->dispose = gtk_source_completion_context_dispose;
 
 	/**
@@ -162,19 +246,6 @@ gtk_source_completion_context_class_init (GtkSourceCompletionContextClass *klass
 	                                                      GTK_TYPE_SOURCE_COMPLETION,
 	                                                      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 
-	/**
-	 * GtkSourceCompletionContext:view:
-	 *
-	 * The #GtkSourceView associated with the context.
-	 **/
-	g_object_class_install_property (object_class,
-	                                 PROP_VIEW,
-	                                 g_param_spec_object ("view",
-	                                                      _("View"),
-	                                                      _("The GtkSourceView"),
-	                                                      GTK_TYPE_SOURCE_VIEW,
-	                                                      G_PARAM_READABLE));
-
 	/**
 	 * GtkSourceCompletionContext:iter:
 	 *
@@ -182,11 +253,12 @@ gtk_source_completion_context_class_init (GtkSourceCompletionContextClass *klass
 	 **/
 	g_object_class_install_property (object_class,
 	                                 PROP_ITER,
-	                                 g_param_spec_pointer ("iter",
+	                                 g_param_spec_boxed ("iter",
 	/* Translators: The GtkTextIter at which the completion was invoked */
-	                                                       _("Iterator"),
-	                                                       _("The GtkTextIter at which the completion was invoked"),
-	                                                       G_PARAM_READWRITE));
+	                                                      _("Iterator"),
+	                                                      _("The GtkTextIter at which the completion was invoked"),
+	                                                      GTK_TYPE_TEXT_ITER,
+	                                                      G_PARAM_READWRITE));
 
 	/**
 	 * GtkSourceCompletionContext:activation:
@@ -206,9 +278,9 @@ gtk_source_completion_context_class_init (GtkSourceCompletionContextClass *klass
 }
 
 static void
-gtk_source_completion_context_init (GtkSourceCompletionContext *self)
+gtk_source_completion_context_init (GtkSourceCompletionContext *context)
 {
-	self->priv = GTK_SOURCE_COMPLETION_CONTEXT_GET_PRIVATE (self);
+	context->priv = GTK_SOURCE_COMPLETION_CONTEXT_GET_PRIVATE (context);
 }
 
 /**
@@ -240,23 +312,6 @@ gtk_source_completion_context_add_proposals (GtkSourceCompletionContext  *contex
 	                                      finished);
 }
 
-/**
- * gtk_source_completion_context_get_view:
- * @context: A #GtkSourceCompletionContext
- * 
- * Get the #GtkSourceView to which the context applies
- *
- * Returns: A #GtkSourceView
- *
- **/
-GtkSourceView *
-gtk_source_completion_context_get_view (GtkSourceCompletionContext *context)
-{
-	g_return_val_if_fail (GTK_IS_SOURCE_COMPLETION_CONTEXT (context), NULL);
-
-	return gtk_source_completion_get_view (context->priv->completion);
-}
-
 /**
  * gtk_source_completion_context_get_iter:
  * @context: A #GtkSourceCompletionContext
@@ -270,9 +325,22 @@ void
 gtk_source_completion_context_get_iter (GtkSourceCompletionContext *context,
                                         GtkTextIter                *iter)
 {
+	GtkTextBuffer *buffer;
+
 	g_return_if_fail (GTK_IS_SOURCE_COMPLETION_CONTEXT (context));
 
-	*iter = context->priv->iter;
+	buffer = get_buffer (context);
+
+	if (context->priv->mark != NULL)
+	{
+		gtk_text_buffer_get_iter_at_mark (buffer, iter, context->priv->mark);
+	}
+	else
+	{
+		/* This should never happen: context should be always be created
+		   providing a position iter */
+		g_warning ("Completion context without mark");
+	}
 }
 
 /**
@@ -298,3 +366,15 @@ _gtk_source_completion_context_cancel (GtkSourceCompletionContext *context)
 
 	g_signal_emit (context, context_signals[CANCELLED], 0);
 }
+
+GtkSourceCompletionContext *
+_gtk_source_completion_context_new (GtkSourceCompletion *completion, GtkTextIter *position)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_COMPLETION (completion), NULL);
+	g_return_val_if_fail (position != NULL, NULL);
+
+	return g_object_new (GTK_TYPE_SOURCE_COMPLETION_CONTEXT,
+	                     "completion", completion,
+	                     "iter", position,
+	                      NULL);
+}
